// ActiveModeTest.cpp : Definiert den Einstiegspunkt für die Konsolenanwendung.
//

#include <memory>
#include <signal.h>
#include "../../api/df_api.h"
#include <map>
#include <vector>
#include <chrono>
#include <thread>
#include <sstream>
#include <cstring>

// Flag zum beenden der Anwendungshauptschleife.
bool g_terminate = false;

enum EnumState
{
    Unknown,
    Connected,
    Alive
};

struct DeviceData
{
    EnumState state;
    std::vector<std::string> infos;
};

void terminate( int unused )
{
    (void)unused;

    int error = 0;
    if ( !DFCStopActiveConnection( &error ) )
    {
        std::printf( "Fehler beim beenden des Active-Mode: %i\n", error );
    }
    else
    {
        std::printf( "Active-Mode erfolgreich beendet.\n");
    }

    // Globales Beenden Flag setzen.
    g_terminate = true;
}

void banner()
{
    std::printf( "Pruefanwendung Active-Mode (Datafox GmbH)\n" );
    std::printf( "Abbruch der Anwendung durch druecken von Strg-C.\n" );
}

void usage()
{
    std::printf( "Usage: ActiveModeTest ip port timeout\n" );
    std::printf( "Default: ActiveModeTest 0.0.0.0 8000 5000\n" );
}

void changeConnectionState( int channelId, EnumState state, const std::vector<std::string> &infos )
{
    switch ( state )
    {
    case Unknown:
        std::printf( "[-] Verbindung getrennt (%i): %s_%s ", channelId, infos[0].c_str(), infos[1].c_str() );
        break;

    case Connected:
        std::printf( "[+] Verbindung hergestellt (%i): %s_%s ", channelId, infos[0].c_str(), infos[1].c_str() );
        break;

    case Alive:
        std::printf( "[i] Verbindung besteht (%i): %s_%s ", channelId, infos[0].c_str(), infos[1].c_str() );
        break;

    default:
        std::printf( "Unbekannter Verbindungszustand." );
        return;
    }

    if ( infos.size() > 2 )
        std::printf( "FW(%s) ", infos[2].c_str() );

    if ( infos.size() > 3 )
        std::printf( "IP:PORT(%s) ", infos[3].c_str() );

    if ( infos.size() > 4 )
        std::printf( "MAC(%s) ", infos[4].c_str() );

    std::printf( "\n" );
}

void printRecord( int channelId, const std::vector<char> &data, std::vector<char>::size_type size )
{
    if ( size > data.size() )
        return;

    bool nulls = false;
    printf( "[%i]: ", channelId );
    auto iter = data.begin();
    for ( std::vector<char>::size_type i = 0; i < size; ++i, ++iter )
    {
        if ( (*iter) == 20 )
        {
            nulls = false;

            // Von einem Datum-Uhrzeitfeld ausgehen.
            printf( "%02i.%02i.%04i %02i:%02i:%02i ",
                *(iter + 3), *(iter + 2), *(iter + 1) + *iter * 100,
                *(iter + 4),  *(iter + 5), *(iter + 6) );

            iter += 6;
            i += 6;
        }
        else
        {
            if ( *iter )
            {
                printf( "%c", *iter );
                nulls = false;
            }
            else
            {
                if ( !nulls )
                {
                    printf( " " );
                    nulls = true; // Mehrere Nullen hintereinander ignorieren.
                }
            }
        }
    }

    printf( "\n" );
}

int main(int argc, char* argv[])
{
    int res;
    int error = 0;
    int channelId;
    int deviceId;
    std::vector<char> ip;
    ip.resize( 64, '\0' );
    std::strcpy( ip.data(), "0.0.0.0" );
    int port = 8000;
    int timeout = 5000;
    std::vector<char> info;
    std::vector<char>::size_type infoSize;
    std::vector<char> buffer;
    std::vector<char>::size_type bufferSize;

    // Größe des Puffers für Datensätze.
    buffer.resize( 260, '\0' );

    // Abfangen des Signals für Ctrl+C
    signal( SIGINT, terminate );
    // Banner
    banner();
    usage();

    //std::printf( "DFComDLL Version %i\n", );

    for ( int i = 1; i < argc; ++i )
    {
        std::printf( "Paramterer %i: %s\n", i, argv[i] );
        switch ( i )
        {
        case 1:
            std::strncpy( ip.data(), argv[i], ip.size() - 1 );
            break;

        case 2:
            port = std::atoi( argv[i] );
            break;

        case 3:
            timeout = std::atoi( argv[i] );
            break;
        }
    }


    std::printf( "Starte Active-Mode mit Parametern: ip:port %s:%i, timeout %i ...", ip.data(), port, timeout );
    // Starten des Active-Mode.
    if ( !DFCStartActiveConnection( ip.data(), port, 0, timeout, 60, 10, 7, 0, &error ) )
    {
        std::printf( "Fehler %i\n", error );
#ifdef _DEBUG
        system( "pause" );
#endif
        return error;
    }
    else
    {
        std::printf( "Ok\n" );
    }

    // Hauptschleife
    std::chrono::milliseconds dura( 200 );
    std::map<int, std::unique_ptr<DeviceData> > mapStates;
    std::map<int, std::unique_ptr<DeviceData>>::iterator iter;

    // Schleife zur Abfrage ob Datensätze vorliegen.
    for ( int loop = 0; !g_terminate; ++loop )
    {
        // Alle aktuellen Zustände auf unbekannt.
        for ( iter = mapStates.begin(); iter != mapStates.end(); ++iter )
            (*iter).second->state = Unknown;

        // Aktuelle Verbindungsmap prüfen.
        if ( (channelId = DFCGetFirstActiveChannelID()) != -1 )
        {
            do
            {
                // Neu oder bereits vorhanden.
                if ( (iter = mapStates.find( channelId )) == mapStates.end() )
                {
                    // Neue Verbindung erkannt.
                    mapStates[channelId] = std::unique_ptr<DeviceData>( new DeviceData );
                    mapStates[channelId]->state = Connected;
                    mapStates[channelId]->infos.clear();
                    info.clear();
                    infoSize = 250;
                    info.resize( infoSize, '\0' );
                    if ( !DFCGetInfoActiveChannel( channelId, info.data(), reinterpret_cast<int*>( &infoSize ) ) )
                    {
                        std::printf( "Fehler bei der Ermittlung von Daten zum Kanal: %i\n", channelId );
                    }
                    else
                    {
                        std::string token;
                        std::stringstream ss( info.data() );
                        while ( std::getline( ss, token, '\r' ) )
                            mapStates[channelId]->infos.push_back( token );

                        // Augabe des Verbindungszustandes.
                        changeConnectionState( channelId, mapStates[channelId]->state, mapStates[channelId]->infos );
                    }
                }
                else
                {
                    // Vorhandene Verbindung beihehalten.
                    (*iter).second->state = Alive;
                }

            } while ( (channelId = DFCGetNextActiveChannelID( channelId )) != -1 );
        }

        std::vector<int> deleteKeys;
        // Alle aktuellen Zustände auf unbekannt.
        for ( iter = mapStates.begin(); iter != mapStates.end(); ++iter )
        {
            if ( (*iter).second->state == Unknown )
            {
                // Augabe des Verbindungszustandes.
                changeConnectionState( (*iter).first, (*iter).second->state, (*iter).second->infos );
                // Für Löschen vormerken.
                deleteKeys.push_back( (*iter).first );
            }
        }

        for ( auto iterKey = deleteKeys.begin(); iterKey != deleteKeys.end(); ++iterKey )
            mapStates.erase( *iterKey );

        // Ggf. auf Datensätze prüfen.
        if ( mapStates.size() )
        {
            info.clear();
            infoSize = 250;
            info.resize( infoSize, '\0' );
            if ( !DFCRecordAvailable( &channelId, &deviceId, info.data(), reinterpret_cast<int *>( &infoSize ), &error ) )
            {
                if ( error == 41 )
                {
                    // Es liegen keine Daten vor.
                }
                else
                {
                    std::printf( "Fehler bei der Abfrage auf Daten. (%i)\n", error );
                }
            }
            else
            {
                for ( ;; )
                {
                    std::fill( buffer.begin(), buffer.end(), '\0' );
                    bufferSize = buffer.size();
                    // Alle Datensätze von diesem Gerät abrufen.
                    if ( (res = DFCReadRecord( channelId, deviceId, buffer.data(), reinterpret_cast<int *>( &bufferSize ), &error )) < 0 )
                    {
                        std::printf( "Fehler beim Abrufen von einem Datensatz. (%i)\n", error );
                        break;
                    }
                    else if ( res > 0 )
                    {
                        // Daten ausgeben.
                        printRecord( channelId, buffer, bufferSize );

                        // Quittieren.
                        if ( DFCQuitRecord( channelId, deviceId, &error ) == -1 )
                        {
                            std::printf( "Fehler beim Quittieren eines Datensatzes. (%i)\n", error );
                            break;
                        }
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        // Kurze Pause
        std::this_thread::sleep_for( dura );
    }

    return 0;
}
